(*
    Id3Tag - ID3 tag editing functions
    Copyright (c) 1998-2002,2003 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Id3;

interface

uses SysUtils, Classes;

const
  MaxGenre = 114;
type
  (* Information about the ID3 tag *)
  TId3InfoFlag = (
    ifHasTag,
    ifTitleChanged,
    ifArtistChanged,
    ifAlbumChanged,
    ifYearChanged,
    ifGenreChanged,
    ifCommentChanged,
    ifTrackChanged
    );
  TId3InfoFlags = set of TId3InfoFlag;
  (* ID3 tag *)
  TId3Info = class
    FTitle: string;
    FArtist: string;
    FAlbum: string;
    FYear: string;
    FGenre: integer;
    FComment: string;
    FTrack: integer;
    FFlags: TId3InfoFlags;
    (* *)
    function GetChanged: boolean;
    procedure SetTitle( Value: string );
    procedure SetArtist( Value: string );
    procedure SetAlbum( Value: string );
    procedure SetYear( Value: string );
    procedure SetGenre( Value: integer );
    procedure SetComment( Value: string );
    procedure SetTrack( Value: integer );
  protected
    procedure Reset;
  public
    constructor Create;
    //destructor Destroy;
    procedure LoadFromFile( Filename: string );
    procedure LoadFromStream( Stream: TStream );
    procedure SaveToFile( Filename: string );
    procedure SaveToStream( Stream: TStream );
    procedure Assign( Value: TId3Info );
    procedure ClearChanged;
    (* *)
    property Title: string read FTitle write SetTitle;
    property Artist: string read FArtist write SetArtist;
    property Album: string read FAlbum write SetAlbum;
    property Year: string read FYear write SetYear;
    property Genre: integer read FGenre write SetGenre;
    property Comment: string read FComment write SetComment;
    property Track: integer read FTrack write SetTrack;
    property Flags: TId3InfoFlags read FFlags;
    property Changed: boolean read GetChanged;
  end;

  function GetGenreName( Index: integer ): string;

implementation

const
  GenreName: array[0..MaxGenre] of PChar = (
  'Blues',
  'Classic Rock',
  'Country',
  'Dance',
  'Disco',
  'Funk',
  'Grunge',
  'Hip-Hop',
  'Jazz',
  'Metal',
  'New Age',
  'Oldies',
  'Other',
  'Pop',
  'R&B',
  'Rap',
  'Reggae',
  'Rock',
  'Techno',
  'Industrial',
  'Alternative',
  'Ska',
  'Death Metal',
  'Pranks',
  'Soundtrack',
  'Euro-Techno',
  'Ambient',
  'Trip-Hop',
  'Vocal',
  'Jazz+Funk',
  'Fusion',
  'Trance',
  'Classical',
  'Instrumental',
  'Acid',
  'House',
  'Game',
  'Sound Clip',
  'Gospel',
  'Noise',
  'AlternRock',
  'Bass',
  'Soul',
  'Punk',
  'Space',
  'Meditative',
  'Instrumental Pop',
  'Instrumental Rock',
  'Ethnic',
  'Gothic',
  'Darkwave',
  'Techno-Industrial',
  'Electronic',
  'Pop-Folk',
  'Eurodance',
  'Dream',
  'Southern Rock',
  'Comedy',
  'Cult',
  'Gangsta',
  'Top 40',
  'Christian Rap',
  'Pop/Funk',
  'Jungle',
  'Native American',
  'Cabaret',
  'New Wave',
  'Psychadelic',
  'Rave',
  'Showtunes',
  'Trailer',
  'Lo-Fi',
  'Tribal',
  'Acid Punk',
  'Acid Jazz',
  'Polka',
  'Retro',
  'Musical',
  'Rock & Roll',
  'Hard Rock',
  'Folk',
  'Folk/Rock',
  'National Folk',
  'Swing',
  'Bebop',
  'Latin',
  'Revival',
  'Celtic',
  'Bluegrass',
  'Avantgarde',
  'Gothic Rock',
  'Progressive Rock',
  'Psychedelic Rock',
  'Symphonic Rock',
  'Slow Rock',
  'Big Band',
  'Chorus',
  'Easy Listening',
  'Acoustic',
  'Humour',
  'Speech',
  'Chanson',
  'Opera',
  'Chamber Music',
  'Sonata',
  'Symphony',
  'Booty Bass',
  'Primus',
  'Porn Groove',
  'Satire',
  'Slow Jam',
  'Club',
  'Tango',
  'Samba',
  'Folklore'
  );

function GetGenreName( Index: integer ): string;
begin
  Result := '';
  if( Index >= 0 )and( Index <= MaxGenre ) then begin
    Result := GenreName[Index];
  end;
end;

constructor TId3Info.Create;
begin
  inherited Create;
  Reset;
end;

procedure TId3Info.Reset;
begin
  FFlags := [];
  FTitle := '';
  FArtist := '';
  FAlbum := '';
  FYear := '';
  FGenre := 255;
  FTrack := 0;
  FComment := '';
end;

procedure TId3Info.SetTitle( Value: string );
begin
  if( Value <> FTitle ) then begin
    FTitle := Value;
    Include( FFlags, ifTitleChanged );
  end;
end;

procedure TId3Info.SetArtist( Value: string );
begin
  Value := Copy( Value, 1, 30 );
  if( Value <> FArtist ) then begin
    FArtist := Value;
    Include( FFlags, ifArtistChanged );
  end;
end;

procedure TId3Info.SetAlbum( Value: string );
begin
  Value := Copy( Value, 1, 30 );
  if( Value <> FAlbum ) then begin
    FAlbum := Value;
    Include( FFlags, ifAlbumChanged );
  end;
end;

procedure TId3Info.SetYear( Value: string );
begin
  Value := Copy( Value, 1, 4 );
  if( Value <> FYear ) then begin
    FYear := Value;
    Include( FFlags, ifYearChanged );
  end;
end;

procedure TId3Info.SetGenre( Value: integer );
begin
  if( Value < 0 )or( Value > MaxGenre ) then
    Value := $FF;
  if( Value <> FGenre ) then begin
    FGenre := Value;
    Include( FFlags, ifGenreChanged );
  end;
end;

procedure TId3Info.SetComment( Value: string );
begin
  Value := Copy( Value, 1, 30 );
  if( Value <> FComment ) then begin
    FComment := Value;
    Include( FFlags, ifCommentChanged );
  end;
end;

procedure TId3Info.SetTrack( Value: integer );
begin
  if( Value < 0 )or( Value > 255 ) then
    Value := 0;
  if( Value <> FTrack ) then begin
    FTrack := Value;
    Include( FFlags, ifTrackChanged );
  end;
end;

type
  TRawId3 = array[ 1..128 ] of char;

function GetRawString( var Buf: TRawId3; Pos, Len: integer ): string;
var
  S: string;
  I: integer;
begin
  S := StringOfChar( ' ', Len );
  I := 1;
  while( I <= Len )and( Buf[I+Pos-1] <> #0 ) do begin
    S[I] := Buf[I+Pos-1];
    Inc( I );
  end;
  Result := Trim( S );
end;

procedure SetRawString( var Buf: TRawId3; Pos, Len: integer; Value: string );
var
  I: integer;
begin
  for I:=1 to Len do begin
    Buf[I+Pos-1] := #0;
    if( I <= Length(Value) ) then
      Buf[I+Pos-1] := Value[I];
  end;
end;

procedure TId3Info.LoadFromStream( Stream: TStream );
var
  Buf: TRawId3;
begin
  try
    Stream.Seek( -sizeof(Buf), soFromEnd );
    Stream.Read( Buf, sizeof(Buf) );
    Reset;
    if( Buf[1] = 'T' )and( Buf[2] = 'A' )and( Buf[3] = 'G' ) then begin
      SetTitle( GetRawString( Buf, 4, 30 ) );
      SetArtist( GetRawString( Buf, 34, 30 ) );
      SetAlbum( GetRawString( Buf, 64, 30 ) );
      SetYear( GetRawString( Buf, 94, 4 ) );
      SetComment( GetRawString( Buf, 98, 30 ) );
      SetGenre( byte( Buf[128] ) );
      if( Buf[126] = #0 ) then
        FTrack := Ord( Buf[127] );
      FFlags := [ifHasTag];
    end;
  except
  end;
end;

function LastPos( C: char; const S: string ): integer;
var
  I: integer;
begin
  Result := Pos( C, S );
  if( Result > 0 ) then begin
    for I:=Length(S) downto Result+1 do begin
      if( S[I] = C ) then begin
        Result := I;
        Exit;
      end;
    end;
  end;
end;

procedure TId3Info.LoadFromFile( Filename: string );
var
  F: TFileStream;
begin
  F := nil;
  try
    F := TFileStream.Create( Filename, fmOpenRead );
    LoadFromStream( F );
{$ifdef VERSION_1_01}
    if not( ifHasTag in Flags ) then begin
      (* Guess at least the title from the filename *)
      Filename := ChangeFileExt( ExtractFilename(Filename), '' );
      for I:=1 to Length(Filename) do
        if( Filename[I] = '_' ) then Filename[I] := ' ';
      I := LastPos( '-', Filename );
      Title := Trim( Copy( Filename, I+1, 30 ) );
    end;
{$endif}
  finally
    F.Free;
  end;
end;

procedure TId3Info.Assign( Value: TId3Info );
begin
  if( Value = nil ) then begin
    Reset;
  end
  else begin
    FFlags := Value.Flags;
    FTitle := Value.Title;
    FArtist := Value.Artist;
    FAlbum := Value.Album;
    FYear := Value.Year;
    FGenre := Value.Genre;
    FComment := Value.Comment;
    FTrack := Value.Track;
  end;
end;

procedure TId3Info.ClearChanged;
begin
  if( ifHasTag in FFlags ) then
    FFlags := [ifHasTag]
  else
    FFlags := [];
end;

procedure TId3Info.SaveToStream( Stream: TStream );
var
  Buf: TRawId3;
begin
  try
    SetRawString( Buf, 1, 3, 'TAG' );
    SetRawString( Buf, 4, 30, FTitle );
    SetRawString( Buf, 34, 30, FArtist );
    SetRawString( Buf, 64, 30, FAlbum );
    SetRawString( Buf, 94, 4, FYear );
    SetRawString( Buf, 98, 30, FComment );
    if( Track <> 0 ) then begin
      Buf[126] := #0;
      Buf[127] := Chr( Track );
    end;
    Buf[128] := char( FGenre );
    if( ifHasTag in FFlags ) then begin
      Stream.Seek( -sizeof(Buf), soFromEnd );
    end
    else begin
      Stream.Seek( 0, soFromEnd );
    end;
    Stream.Write( Buf, sizeof(Buf) );
    Include( FFlags, ifHasTag );
  except
  end;
end;

procedure TId3Info.SaveToFile( Filename: string );
var
  F: TFileStream;
begin
  F := nil;
  try
    F := TFileStream.Create( Filename, fmOpenReadWrite );
    SaveToStream( F );
  finally
    F.Free;
  end;
end;

function TId3Info.GetChanged: boolean;
begin
  Result := [ifTitleChanged,ifArtistChanged,ifAlbumChanged,ifYearChanged,ifGenreChanged,ifCommentChanged,ifTrackChanged] * FFlags <> [];
end;

end.
